import numpy as np
import matplotlib.pyplot as plt
import japanize_matplotlib


class PlotPointOnCircle:
    """
    座標平面上に点P(x, y)とその射影，角度，ラベルを描画するクラス
    """
    def __init__(self, radius, theta_deg):
        """
        円と三角形の描画に必要なデータを初期化
        :param radius: 円の半径
        :param theta_deg: 中心角の角度（度）
        """
        self._r = radius
        self._theta_deg = theta_deg
        self._theta = np.deg2rad(theta_deg)  # ラジアンに変換

        # ラベル関係パラメータ設定
        config = self._configure_labels()  # ラベル位置の微調整パラメータ
        self._ha_pos1, self._ha_pos2, self._va_pos1, self._va_pos2, self._ha_sign, self._va_sign \
             = config.values()
        self._labelsize = 16 # 頂点，辺のラベルサイズ
        self._angle_labelsize = 12 # 角度のラベルサイズ
        self._pos_offset5 = 0.05 * self._r # ラベル位置の微調整パラメータ(円の半径に応じて調整幅を変更)
        self._pos_offset3 = 0.03 * self._r
        
        # 図形関係パラメータ設定
        self._circle_linewidth = 2 # 半径の線幅
        self._angle_linewidth = 1 # 円弧の線幅
        self._proj_linewidth = 1 # 射影線の線幅
        self._ptsize = 100 # 円上の点のサイズ

    def validate_angle180(self):
        """角度の範囲を確認"""
        plt.ylim(-self._r * 0.1, self._r * 1.2) # 縦方向の描画範囲を設定
        if self._theta_deg < 0 or self._theta_deg > 180:
            raise ValueError("角度は 0° 以上 180° 以下で指定してください。")

    def validate_angle360(self):
        """角度の範囲を確認"""
        plt.ylim(-self._r * 1.2, self._r * 1.2) # 縦方向の描画範囲を設定
        if self._theta_deg < -180 or self._theta_deg > 180:
            raise ValueError("角度は -180° 以上 180° 以下で指定してください。")

    def draw_point(self):
        """円上に点Pをプロット，ラベルP(x, y)を印字"""
        plt.scatter(self._r * np.cos(self._theta), self._r * np.sin(self._theta), color='red', s=self._ptsize, zorder = 5)
        
        # 点Pのラベル
        plt.text(self._r * np.cos(self._theta), self._r * np.sin(self._theta) + np.sign(self._theta_deg) * self._pos_offset5,
                 r'$P(x, y)$', fontsize=self._labelsize, ha=self._ha_pos1, va=self._va_pos1)

    def draw_radius(self):
        """半径の線を描画"""
        plt.plot([0, self._r * np.cos(self._theta)], [0, self._r * np.sin(self._theta)],
                 color='blue', linewidth=self._circle_linewidth, label=r'半径 ($r$)')

    def _configure_labels(self):
        """ラベル位置の設定"""
        if self._theta_deg > 90:
            return {'ha_pos1': 'right', 'ha_pos2': 'left', 'va_pos1': 'bottom', 'va_pos2': 'top', 'ha_sign': 1, 'va_sign': 1}
        elif self._theta_deg >= 0:
            return {'ha_pos1': 'left', 'ha_pos2': 'right', 'va_pos1': 'bottom', 'va_pos2': 'top', 'ha_sign': -1, 'va_sign': 1}
        elif self._theta_deg > -90:
            return {'ha_pos1': 'left', 'ha_pos2': 'right', 'va_pos1': 'top', 'va_pos2': 'bottom', 'ha_sign': -1, 'va_sign': -1}
        else:
            return {'ha_pos1': 'right', 'ha_pos2': 'left', 'va_pos1': 'top', 'va_pos2': 'bottom', 'ha_sign': 1, 'va_sign': -1}
            
    def draw_projections(self):
        """x軸およびy軸への射影線を描画"""
        plt.plot([self._r * np.cos(self._theta), self._r * np.cos(self._theta)],
                 [0, self._r * np.sin(self._theta)], color='red', linestyle='--', linewidth=self._proj_linewidth)
        plt.plot([0, self._r * np.cos(self._theta)],
                 [self._r * np.sin(self._theta), self._r * np.sin(self._theta)], color='green', linestyle='--', linewidth=self._proj_linewidth)
        
    def draw_arc_and_labels(self):
        """角度を示す円弧を描画し、ラベルを印字"""
        # 円弧の描画
        theta_arc = np.linspace(0, self._theta, 100)
        x_arc = 0.25 * self._r * np.cos(theta_arc)
        y_arc = 0.25 * self._r * np.sin(theta_arc)
        plt.plot(x_arc, y_arc, color='black', linewidth = self._angle_linewidth)

        # 角度 θ のラベル
        mid_arc_x = 0.35 * self._r * np.cos(self._theta / 2)
        mid_arc_y = 0.35 * self._r * np.sin(self._theta / 2)
        plt.text(mid_arc_x - self._pos_offset5, mid_arc_y - self._pos_offset5, rf'$\theta = {self._theta_deg}$°', fontsize=self._angle_labelsize)

        # 半径ラベル
        plt.text(self._r * np.cos(self._theta) / 2 + self._ha_sign * self._pos_offset5, \
                 self._r * np.sin(self._theta) / 2 + self._va_sign * self._pos_offset5, \
                 r'$r$', fontsize=self._labelsize, color='blue', ha=self._ha_pos2, va='center')

    def draw_labels(self):
        """射影値x, yと原点Oのラベルを描画"""
        # 射影値xとyのラベル
        plt.text(self._r * np.cos(self._theta), -self._va_sign * self._pos_offset3, r'$x$', fontsize=16, ha='center', va=self._va_pos2)
        plt.text(self._ha_sign * self._pos_offset3, self._r * np.sin(self._theta), r'$y$', fontsize=16, ha=self._ha_pos2, va='center')

        # 原点Oのラベル
        plt.text(-self._pos_offset3, -self._va_sign * self._pos_offset3, r'$O$', fontsize=self._labelsize, ha='right', va=self._va_pos2)


class CirclePlotAuxiliary:
    def __init__(self, radius):

        self._r = radius

        # 図形関係パラメータ設定
        self._circle_linewidth = 2 # 円の線幅
        self._axis_linewidth = 1 # 軸の線幅
        
        # プロットのパラメータ設定
        self.fig, self.ax = plt.subplots(figsize=(5, 5)) # 描画範囲 単位はインチ
        self.ax.set_aspect('equal')  # 図形の縦横比を1:1に保ちながら描画範囲に収まるように描画する
    
    def draw_full_arc(self):
        """円の上半分を描画"""
        theta_full_arc = np.linspace(0, 2*np.pi, 200)
        x_full_arc = self._r * np.cos(theta_full_arc)
        y_full_arc = self._r * np.sin(theta_full_arc)
        plt.plot(x_full_arc, y_full_arc, color='black', linestyle='--', linewidth = self._circle_linewidth)

    def draw_axes(self):
        """x軸とy軸を描画"""
        plt.axhline(0, color='black', linewidth=self._axis_linewidth)
        plt.axvline(0, color='black', linewidth=self._axis_linewidth)

    def finalize_plot(self):
        """プロットの範囲を設定し、装飾を仕上げる"""
        plt.gca().set_aspect('equal', adjustable='box')
        plt.xlim(-self._r * 1.2, self._r * 1.2) # 描画範囲を設定  ylimはvalidate_angle180, validate_angle360 メソッド内で設定        
        plt.xticks([])
        plt.yticks([])
        for spine in ['left', 'bottom', 'right', 'top']:
            plt.gca().spines[spine].set_visible(False)






