import numpy as np
import matplotlib.pyplot as plt

class Triangle_Base:
    """ 
    三角形のみを描画するクラス
    頂点ABCの配置が時計回りか反時計回りかを判定し，時計回りの場合はBとCの順番を入れ換える
    """
    def __init__(self, A, B, C):
        """ 頂点ABCの配置が時計回りか反時計回りかを判定する．
        #　　全てのコードは，A→B→Cが反時計回りであることを前提に書かれている．
        #  　A→B→Cが時計回りで与えられた場合は，BとCの情報を入れ換えて，反時計回りの順番とする """
        self.X, self.Y, self.Z, \
        self.labelX, self.labelY, self.labelZ, \
        self.label_x, self.label_y, self.label_z \
            = self.vertex_sequence(A, B, C)

        # 三角形の描画
        triangle_points = np.array([self.X, self.Y, self.Z, self.X])
        plt.plot(triangle_points[:, 0], triangle_points[:, 1], 'b-', label='三角形ABC', linewidth=2)
        
    def vertex_sequence(self, A, B, C):
        """
        頂点 A, B, C の順序を反時計回りに整列し、対応するラベルを設定

        Returns:
            tuple: 整列された頂点 X, Y, Z とラベル ('A', 'B', 'C', 'a', 'b', 'c')
        """
        start_angle, end_angle = self._peak_search_fn(A, B, C)
        # 順序を反時計回りに整列
        if end_angle > start_angle:
            X, Y, Z = A, B, C
            labels = ('A', 'B', 'C', 'a', 'b', 'c')
        else:
            X, Y, Z = A, C, B
            labels = ('A', 'C', 'B', 'a', 'c', 'b')

        # 整列された頂点とラベルを返す
        return X, Y, Z, *labels

    def _peak_search_fn(self, A, B, C):
        """ 最高位置にある頂点を探索して，その頂点を挟む辺の角度を求める関数 """
            # 三角形の一番高い頂点からそれより低い位置にある他の２頂点との角度を求めれば，（arctan2の符号が反転する）180°を後の判定対象から外せる
            # ただし，最高位の頂点が２つある場合は，左側の頂点を選択することで，180°を避けられる．
        if A[1] >= B[1] and A[1] > C[1]:
            X, Y, Z = (A, B, C)
            if A[1] == B[1] and B[0] < A[0]:
                X, Y, Z = (B, C, A)
        elif B[1] >= C[1] and B[1] > A[1]:
            X, Y, Z = (B, C, A)
            if B[1] == C[1] and C[0] < B[0]:
                X, Y, Z = (C, A, B)            
        elif C[1] >= A[1] and C[1] > B[1]:
            X, Y, Z = (C, A, B)
            if C[1] == A[1] and A[0] < C[0]:
                X, Y, Z = (A, B, C)
        
        start_angle = np.arctan2(Y[1] - X[1], Y[0] - X[0]) # XからYを見る角度　右水平方向を 0 として，反時計回りを正とする．-π < start_angle < π
        end_angle = np.arctan2(Z[1] - X[1], Z[0] - X[0])
        return start_angle, end_angle


class Circumcircle:
    """
    外接円描画の機能を提供するクラス
    """
    def __init__(self, A, B, C):
        """外接円の中心と半径を計算"""
        self._center, self._radius = self._calculate_circumcircle(A, B, C)

        """ パラメータ設定 """
        self._labelsize = 18                  # ラベルのサイズ

    @property
    def radius(self):
        """外接円の半径を取得"""
        return self._radius    

    @property
    def center(self):
        """外心を取得"""
        return self._center
    
    def _calculate_circumcircle(self, A, B, C):
        """外接円の中心と半径を計算"""
        D = 2 * (A[0] * (B[1] - C[1]) + B[0] * (C[1] - A[1]) + C[0] * (A[1] - B[1]))
        Ux = ((A[0]**2 + A[1]**2) * (B[1] - C[1]) +
              (B[0]**2 + B[1]**2) * (C[1] - A[1]) +
              (C[0]**2 + C[1]**2) * (A[1] - B[1])) / D
        Uy = ((A[0]**2 + A[1]**2) * (C[0] - B[0]) +
              (B[0]**2 + B[1]**2) * (A[0] - C[0]) +
              (C[0]**2 + C[1]**2) * (B[0] - A[0])) / D
        center = np.array([Ux, Uy])
        radius = np.linalg.norm(center - A)
        return center, radius

    def draw_circumcircle(self):
        """外接円を描画"""
        theta = np.linspace(0, 2 * np.pi, 100)
        circle_x = self._center[0] + self._radius * np.cos(theta)
        circle_y = self._center[1] + self._radius * np.sin(theta)
        plt.plot(circle_x, circle_y, 'r--', label='外接円')

    def draw_circumcenter(self, label):
        """外心を描画"""
        plt.scatter(*self._center, color='green', label='外心')
        plt.text(self._center[0], self._center[1] + 0.05 * self._radius, label, fontsize=self._labelsize, ha='center', va='bottom')


class Incircle:
    """
    内接円描画の機能を提供するクラス
    """
    def __init__(self, A, B, C):
        """内接円の中心と半径を計算"""
        self._incenter, self._inradius = self._calculate_incircle(A, B, C)

        """ パラメータ設定 """
        self._labelsize = 18                  # ラベルのサイズ
        
    def _calculate_incircle(self, A, B, C):
        """内接円の中心と半径を計算"""
        # 辺の長さを計算
        a = np.sqrt((B[0] - C[0])**2 + (B[1] - C[1])**2)
        b = np.sqrt((C[0] - A[0])**2 + (C[1] - A[1])**2)
        c = np.sqrt((A[0] - B[0])**2 + (A[1] - B[1])**2)
        # 周の長さ
        s = (a + b + c) / 2
        # 面積の計算
        area = np.sqrt(s * (s - a) * (s - b) * (s - c))
        # 内心の座標
        incenter_x = (a * A[0] + b * B[0] + c * C[0]) / (a + b + c)
        incenter_y = (a * A[1] + b * B[1] + c * C[1]) / (a + b + c)
        incenter = np.array([incenter_x, incenter_y])
        # 内接円の半径
        inradius = area / s
        return incenter, inradius

    def draw_incircle(self):
        """内接円を描画"""
        theta = np.linspace(0, 2 * np.pi, 100)
        incircle_x = self._incenter[0] + self._inradius * np.cos(theta)
        incircle_y = self._incenter[1] + self._inradius * np.sin(theta)
        plt.plot(incircle_x, incircle_y, 'r--', label='内接円')

    def draw_incenter(self, label):
        """内心の描画と印字"""
        plt.scatter(*self._incenter, color='blue', label='内心')
        plt.text(self._incenter[0], self._incenter[1] - 0.1 * self._inradius, label, fontsize=self._labelsize, ha='right', va='top')

    
class TrianglePlotAuxiliary:
    """
    三角形描画の補助機能を提供するクラス
    """
    def __init__(self, plot_size):
        """ 各種パラメータ設定 """
        self._radius_arc = 0.25 * plot_size      # 円弧半径
        self._linewidth = 1                     # 円弧，直角マークの線幅
        self._labelsize = 20                    # 頂点，辺の長さなどのラベルのサイズ
        self._shift_length = 0.08 * plot_size   # ラベルの印字位置微調整用パラメータ
        self._mark_size = 0.15 * plot_size      # 直角マークのサイズ
        self._shift_angle_label = 1.2 * self._radius_arc # 頂点の角度ラベルの印字位置をradius_arc の1.2倍にする

    def _calculate_angle(self, X, Y, Z):
        """
        頂点 X を基準とした Y と Z の方向角を計算

        Parameters:
            X, Y, Z (tuple): 頂点 X, Y, Z の座標 (x, y)

        Returns:
            tuple: 開始角度 (start_angle) と終了角度 (end_angle) のペア (ラジアン単位)
        """
        # Xから見たYの方向角Yをstart_angle, Xから見たZの方向角をene_angle
        start_angle = np.arctan2(Y[1] - X[1], Y[0] - X[0])
        end_angle = np.arctan2(Z[1] - X[1], Z[0] - X[0])

        # 角度が負の場合の補正
        if start_angle > end_angle:
            end_angle += 2 * np.pi

        return start_angle, end_angle

    def draw_each_arc(self, X, Y, Z):
        """
        頂点 X を中心とした円弧を描画

        Parameters:
            X, Y, Z (tuple): 頂点 X, Y, Z の座標 (x, y)
        """
        # 円弧の開始角度と終了角度を計算
        start_angle, end_angle = self._calculate_angle(X, Y, Z)
        # linespace関数は小さい方の角度から大きい方の角度へと反時計回りに円弧を描く
        if abs(start_angle - end_angle) >= np.pi:
            start_angle = start_angle + 2*np.pi
        if round(abs(start_angle - end_angle)*180/np.pi, 7) == 90:
            # 直角マークの描画
            self.draw_right_angle_mark(X, Y, Z)
        else:
            # 円弧の点の座標を計算
            arc_x = X[0] + self._radius_arc * np.cos(np.linspace(start_angle, end_angle, 100)) 
            arc_y = X[1] + self._radius_arc * np.sin(np.linspace(start_angle, end_angle, 100))
            # 円弧を描画
            plt.plot(arc_x, arc_y, color='black', linewidth=self._linewidth, label='Arc')

    
    def draw_vertex_label(self, X, Y, Z, label):
        """
        頂点 X のラベルを描画

        Parameters:
            X, Y, Z (tuple): 頂点 X, Y, Z の座標 (x, y)
            label (str): 描画するラベルテキスト
        """
        # 頂点 X の角度範囲を計算
        start_angle, end_angle = self._calculate_angle(X, Y, Z)

        # 中間角度を計算
        mid_angle = (start_angle + end_angle) / 2

        # ラベルの位置を計算
        label_x = X[0] - 1.2*self._shift_length * np.cos(mid_angle)
        label_y = X[1] - 1.2*self._shift_length * np.sin(mid_angle)

        # ラベルを描画
        plt.text(
            label_x, label_y, label,
            fontsize=self._labelsize,
            ha='center',
            va='center'
        )

    def draw_side_label(self, X, Y, Z, label1, label2):
        """
        辺 YZ のラベルを描画

        Parameters:
            X, Y, Z (tuple): 頂点 X, Y, Z の座標 (x, y)
        """
        # 角度が90°の場合はラベルを「斜辺」に変更
        start_angle, end_angle = self._calculate_angle(X, Y, Z)
        if round((end_angle - start_angle)*180/np.pi) == 90:
            label1 = '斜辺'

        # 辺 YZ の角度に応じて印字の水平起点を調整
        angleYZ = np.arctan2(Y[1] - Z[1], Y[0] - Z[0])
        if abs(angleYZ) >  35*np.pi/36 or abs(angleYZ) < np.pi/36:
            ha_pos, ratio = 'center', 1.3
        elif angleYZ < 0: 
            ha_pos, ratio = 'left', 1
        else:
            ha_pos, ratio = 'right', 1

        # 辺のラベルを印字
        plt.text((Y[0]+Z[0])/2 - ratio * self._shift_length*np.sin(angleYZ), \
                 (Y[1]+Z[1])/2 + ratio * self._shift_length*np.cos(angleYZ), \
                 rf'{label1}{label2}', fontsize=self._labelsize, color="black", ha=ha_pos, va = 'center')


    def draw_side_length(self, X, Y, Z):
        """
        頂点 Y と Z の間の辺の長さを印字

        Parameters:
            X, Y, Z (tuple): 頂点 X, Y, Z の座標 (x, y)
        """
        # 辺 YZ の角度を計算
        angleYZ = np.arctan2(Y[1] - Z[1], Y[0] - Z[0])

        # 辺 YZ の角度に応じて印字の水平起点を調整
        if abs(angleYZ) >  35*np.pi/36 or abs(angleYZ) < np.pi/36:
            ha_pos, ratio = 'center', 1.3
        elif angleYZ < 0: 
            ha_pos, ratio = 'left', 1
        else:
            ha_pos, ratio = 'right', 1

        # 辺 YZ の長さを計算
        lengthYZ = round(np.sqrt((Y[0] - Z[0])**2 + (Y[1] - Z[1])**2), 2)

        # テキストの印字位置を計算
        text_x = (Y[0] + Z[0]) / 2 - self._shift_length * np.sin(angleYZ)
        text_y = (Y[1] + Z[1]) / 2 + self._shift_length * np.cos(angleYZ)

        # 辺の長さを印字
        plt.text(
            text_x, text_y, lengthYZ,
            fontsize=self._labelsize,
            color="black",
            ha=ha_pos,
            va='center'
        )

    def cal_vertex_angle(self, X, Y, Z):
        """
        頂点 X の角度 θ を計算

        Parameters:
            X, Y, Z (tuple): 頂点 X, Y, Z の座標 (x, y)

        Returns:
            float: 頂点 X の角度 θ（度単位）
        """
        # 頂点 X の角度範囲を計算
        start_angle, end_angle = self._calculate_angle(X, Y, Z)

        # 角度を計算して度単位に変換
        angleX = round((end_angle - start_angle) * 180 / np.pi)

        return angleX

    
    def draw_angle_label(self, X, Y, Z, label):
        """
        頂点 X の角度ラベルを円弧の内側に追加

        Parameters:
            X, Y, Z (tuple): 頂点 X, Y, Z の座標 (x, y)
            label (str): ラベルテキスト
        """
        # 頂点 X の角度の開始角度と終了角度を計算
        start_angle, end_angle = self._calculate_angle(X, Y, Z)
        mid_angle = (start_angle + end_angle) / 2

        # 角度に応じたラベルの配置設定
        if abs(mid_angle) > 3 * np.pi / 4:  # ラベルが円弧の左側に来る場合
            ha_pos = 'right'
            ratio = 0.95*self._shift_angle_label
        elif -np.pi / 4 < mid_angle < np.pi / 4:  # ラベルが円弧の右側に来る場合
            ha_pos = 'left'
            ratio = 0.95*self._shift_angle_label
        else:  # ラベルが円弧の中央付近に来る場合
            ha_pos = 'center'
            ratio = 1.2 * self._shift_angle_label

        # ラベルの印字位置を計算
        label_x = X[0] + ratio * np.cos(mid_angle)
        label_y = X[1] + ratio * np.sin(mid_angle)

        # ラベルを印字
        plt.text(
            label_x, label_y, label,
            fontsize=self._labelsize,
            ha=ha_pos,
            va='center'
        )


    def _draw_right_angle_mark_fn(self, point, direction):
        """
        直角マークを描画する補助関数

        Parameters:
            point (tuple): 直角マークの基準点 (x, y)
            direction (numpy.ndarray): 直角を作るベクトルの方向
        """
        # direction ベクトルを正規化
        direction = direction / np.linalg.norm(direction)

        # 直角マークの頂点を計算
        p1 = point + self._mark_size * np.array([-direction[1], direction[0]])
        p2 = p1 + self._mark_size * direction
        p3 = point + self._mark_size * direction

        # 直角マークを描画
        plt.plot([p1[0], p2[0]], [p1[1], p2[1]], 'k-')
        plt.plot([p2[0], p3[0]], [p2[1], p3[1]], 'k-')

    def draw_right_angle_mark(self, X, Y, Z):
        """
        頂点 X の直角マークを描画

        Parameters:
            X, Y, Z (tuple): 頂点 X, Y, Z の座標 (x, y)
        """
        # 頂点 X の角度を計算
        start_angle, end_angle = self._calculate_angle(X, Y, Z)

        # 角度が 90 度の場合のみ直角マークを描画
        if round((end_angle - start_angle) * 180 / np.pi, 4) == 90:
            XY_direction = np.array([Y[0] - X[0], Y[1] - X[1]])
            self._draw_right_angle_mark_fn(X, XY_direction)

    def draw_center_angle(self, center, B, C):
        """弧BCの中心角とその角を示す円弧を描画"""
        # 外心と頂点を結ぶ線を描画
        plt.plot([center[0], B[0]], [center[1], B[1]], 'k-') # 中心角の描画
        plt.plot([center[0], C[0]], [center[1], C[1]], 'k-')

        # 中心角の円弧を描画（鈍角も描画できる）
        point, next_point, prev_point = (center, B, C)
        start_angle = np.arctan2(next_point[1] - point[1], next_point[0] - point[0])
        end_angle = np.arctan2(prev_point[1] - point[1], prev_point[0] - point[0])
        if start_angle > end_angle:
            end_angle = end_angle + 2*np.pi
        theta = np.linspace(start_angle, end_angle, 100)
        arc_x = point[0] + self._radius_arc*np.cos(theta)
        arc_y = point[1] + self._radius_arc*np.sin(theta)
        plt.plot(arc_x, arc_y, 'black', linewidth = 1)

    def draw_perp_line(self, A, B, C):
        """ 頂点Aから辺BCへの垂線を描画 """ 
        BC_direction = np.array([C[0] - B[0], C[1] - B[1]])
        BC_direction = BC_direction / np.linalg.norm(BC_direction)
        projection_length = np.dot(np.array(A) - np.array(B), BC_direction)
        D = B + projection_length * BC_direction
        plt.plot([A[0], D[0]], [A[1], D[1]], 'k--')  # 垂線を描く

        """ CからDへ辺BCを延長 """
        plt.plot([C[0], D[0]], [C[1], D[1]], 'k--')  

        return D, projection_length

    def sideBC_labelling(self, D, B, C, projection_length):
        """ 辺BCのラベル付け """ 
        # 頂点Aからの垂線と辺BCの交点が辺BC内にあるかを判定 
        if projection_length < 0:
            return D, B, B, C, 'd', 'a'
        elif (projection_length >= 0) and (projection_length <= np.sqrt((B[0] - C[0])**2 + (B[1] - C[1])**2)):
            return B, D, D, C, 'd', 'a-d'
        else:
            return B, C, C, D, 'a', 'd'
            
       